# Functions/classes for WCSAxes related to APE14 WCSes

import numpy as np

from astropy import units as u
from astropy.coordinates import ICRS, BaseCoordinateFrame, SkyCoord
from astropy.wcs import WCS
from astropy.wcs.utils import local_partial_pixel_derivatives
from astropy.wcs.wcsapi import SlicedLowLevelWCS

from .frame import EllipticalFrame, RectangularFrame, RectangularFrame1D
from .transforms import CurvedTransform

__all__ = [
    "transform_coord_meta_from_wcs",
    "WCSWorld2PixelTransform",
    "WCSPixel2WorldTransform",
]

IDENTITY = WCS(naxis=2)
IDENTITY.wcs.ctype = ["X", "Y"]
IDENTITY.wcs.crval = [0.0, 0.0]
IDENTITY.wcs.crpix = [1.0, 1.0]
IDENTITY.wcs.cdelt = [1.0, 1.0]


def transform_coord_meta_from_wcs(wcs, frame_class, slices=None):
    if slices is not None:
        slices = tuple(slices)

    if wcs.pixel_n_dim > 2:
        if slices is None:
            raise ValueError(
                "WCS has more than 2 pixel dimensions, so 'slices' should be set"
            )
        elif len(slices) != wcs.pixel_n_dim:
            raise ValueError(
                "'slices' should have as many elements as WCS "
                "has pixel dimensions (should be {})".format(wcs.pixel_n_dim)
            )

    is_fits_wcs = isinstance(wcs, WCS) or (
        isinstance(wcs, SlicedLowLevelWCS) and isinstance(wcs._wcs, WCS)
    )

    coord_meta = {}
    coord_meta["name"] = []
    coord_meta["type"] = []
    coord_meta["wrap"] = []
    coord_meta["unit"] = []
    coord_meta["visible"] = []
    coord_meta["format_unit"] = []

    for idx in range(wcs.world_n_dim):
        axis_type = wcs.world_axis_physical_types[idx]
        axis_unit = u.Unit(wcs.world_axis_units[idx])
        coord_wrap = None
        format_unit = axis_unit

        coord_type = "scalar"

        if axis_type is not None:
            axis_type_split = axis_type.split(".")

            if "pos.helioprojective.lon" in axis_type:
                coord_wrap = 180.0 * u.deg
                format_unit = u.arcsec
                coord_type = "longitude"
            elif "pos.helioprojective.lat" in axis_type:
                format_unit = u.arcsec
                coord_type = "latitude"
            elif "pos.heliographic.stonyhurst.lon" in axis_type:
                coord_wrap = 180.0 * u.deg
                format_unit = u.deg
                coord_type = "longitude"
            elif "pos.heliographic.stonyhurst.lat" in axis_type:
                format_unit = u.deg
                coord_type = "latitude"
            elif "pos.heliographic.carrington.lon" in axis_type:
                coord_wrap = 360.0 * u.deg
                format_unit = u.deg
                coord_type = "longitude"
            elif "pos.heliographic.carrington.lat" in axis_type:
                format_unit = u.deg
                coord_type = "latitude"
            elif "pos" in axis_type_split:
                if "lon" in axis_type_split:
                    coord_type = "longitude"
                elif "lat" in axis_type_split:
                    coord_type = "latitude"
                elif "ra" in axis_type_split:
                    coord_type = "longitude"
                    format_unit = u.hourangle
                elif "dec" in axis_type_split:
                    coord_type = "latitude"
                elif "alt" in axis_type_split:
                    coord_type = "longitude"
                elif "az" in axis_type_split:
                    coord_type = "latitude"
                elif "long" in axis_type_split:
                    coord_type = "longitude"

        coord_meta["type"].append(coord_type)
        coord_meta["wrap"].append(coord_wrap)
        coord_meta["format_unit"].append(format_unit)
        coord_meta["unit"].append(axis_unit)

        # For FITS-WCS, for backward-compatibility, we need to make sure that we
        # provide aliases based on CTYPE for the name.
        if is_fits_wcs:
            name = []
            if isinstance(wcs, WCS):
                name.append(wcs.wcs.ctype[idx].lower())
                name.append(wcs.wcs.ctype[idx][:4].replace("-", "").lower())
            elif isinstance(wcs, SlicedLowLevelWCS):
                name.append(wcs._wcs.wcs.ctype[wcs._world_keep[idx]].lower())
                name.append(
                    wcs._wcs.wcs.ctype[wcs._world_keep[idx]][:4]
                    .replace("-", "")
                    .lower()
                )
            if name[0] == name[1]:
                name = name[0:1]
            if axis_type:
                if axis_type not in name:
                    name.insert(0, axis_type)
            if wcs.world_axis_names and wcs.world_axis_names[idx]:
                if wcs.world_axis_names[idx] not in name:
                    name.append(wcs.world_axis_names[idx])
            name = tuple(name) if len(name) > 1 else name[0]
        else:
            name = axis_type or ""
            if wcs.world_axis_names:
                name = (
                    (name, wcs.world_axis_names[idx])
                    if wcs.world_axis_names[idx]
                    else name
                )

        coord_meta["name"].append(name)

    coord_meta["default_axislabel_position"] = [""] * wcs.world_n_dim
    coord_meta["default_ticklabel_position"] = [""] * wcs.world_n_dim
    coord_meta["default_ticks_position"] = [""] * wcs.world_n_dim
    # If the world axis has a name use it, else display the world axis physical type.
    fallback_labels = [
        name[0] if isinstance(name, (list, tuple)) else name
        for name in coord_meta["name"]
    ]
    coord_meta["default_axis_label"] = [
        wcs.world_axis_names[i] or fallback_label
        for i, fallback_label in enumerate(fallback_labels)
    ]

    transform_wcs, invert_xy, world_map = apply_slices(wcs, slices)

    transform = WCSPixel2WorldTransform(transform_wcs, invert_xy=invert_xy)

    for i in range(len(coord_meta["type"])):
        coord_meta["visible"].append(i in world_map)

    inv_all_corr = [False] * wcs.world_n_dim
    m = transform_wcs.axis_correlation_matrix.copy()
    if invert_xy:
        inv_all_corr = np.all(m, axis=1)
        m = m[:, ::-1]

    if frame_class is RectangularFrame:
        for i, spine_name in enumerate("bltr"):
            pos = np.nonzero(m[:, i % 2])[0]
            # If all the axes we have are correlated with each other and we
            # have inverted the axes, then we need to reverse the index so we
            # put the 'y' on the left.
            if inv_all_corr[i % 2]:
                pos = pos[::-1]

            if len(pos) > 0:
                index = world_map[pos[0]]
                coord_meta["default_axislabel_position"][index] = spine_name
                coord_meta["default_ticklabel_position"][index] = spine_name
                coord_meta["default_ticks_position"][index] = spine_name
                m[pos[0], :] = 0

        # In the special and common case where the frame is rectangular and
        # we are dealing with 2-d WCS (after slicing), we show all ticks on
        # all axes for backward-compatibility.
        if len(world_map) == 2:
            for index in world_map:
                coord_meta["default_ticks_position"][index] = "bltr"

    elif frame_class is RectangularFrame1D:
        derivs = np.abs(
            local_partial_pixel_derivatives(
                transform_wcs,
                *[0] * transform_wcs.pixel_n_dim,
                normalize_by_world=False,
            )
        )[:, 0]
        for i, spine_name in enumerate("bt"):
            # Here we are iterating over the correlated axes in world axis order.
            # We want to sort the correlated axes by their partial derivatives,
            # so we put the most rapidly changing world axis on the bottom.
            pos = np.nonzero(m[:, 0])[0]
            order = np.argsort(derivs[pos])[::-1]  # Sort largest to smallest
            pos = pos[order]
            if len(pos) > 0:
                index = world_map[pos[0]]
                coord_meta["default_axislabel_position"][index] = spine_name
                coord_meta["default_ticklabel_position"][index] = spine_name
                coord_meta["default_ticks_position"][index] = spine_name
                m[pos[0], :] = 0

        # In the special and common case where the frame is rectangular and
        # we are dealing with 2-d WCS (after slicing), we show all ticks on
        # all axes for backward-compatibility.
        if len(world_map) == 1:
            for index in world_map:
                coord_meta["default_ticks_position"][index] = "bt"

    elif frame_class is EllipticalFrame:
        if "longitude" in coord_meta["type"]:
            lon_idx = coord_meta["type"].index("longitude")
            coord_meta["default_axislabel_position"][lon_idx] = "h"
            coord_meta["default_ticklabel_position"][lon_idx] = "h"
            coord_meta["default_ticks_position"][lon_idx] = "h"

        if "latitude" in coord_meta["type"]:
            lat_idx = coord_meta["type"].index("latitude")
            coord_meta["default_axislabel_position"][lat_idx] = "c"
            coord_meta["default_ticklabel_position"][lat_idx] = "c"
            coord_meta["default_ticks_position"][lat_idx] = "c"

    else:
        for index in range(len(coord_meta["type"])):
            if index in world_map:
                coord_meta["default_axislabel_position"][
                    index
                ] = frame_class.spine_names
                coord_meta["default_ticklabel_position"][
                    index
                ] = frame_class.spine_names
                coord_meta["default_ticks_position"][index] = frame_class.spine_names

    return transform, coord_meta


def apply_slices(wcs, slices):
    """
    Take the input WCS and slices and return a sliced WCS for the transform and
    a mapping of world axes in the sliced WCS to the input WCS.
    """
    if isinstance(wcs, SlicedLowLevelWCS):
        world_keep = list(wcs._world_keep)
    else:
        world_keep = list(range(wcs.world_n_dim))

    # world_map is the index of the world axis in the input WCS for a given
    # axis in the transform_wcs
    world_map = list(range(wcs.world_n_dim))
    transform_wcs = wcs
    invert_xy = False
    if slices is not None:
        wcs_slice = list(slices)
        wcs_slice[wcs_slice.index("x")] = slice(None)
        if "y" in slices:
            wcs_slice[wcs_slice.index("y")] = slice(None)
            invert_xy = slices.index("x") > slices.index("y")

        transform_wcs = SlicedLowLevelWCS(wcs, wcs_slice[::-1])
        world_map = tuple(world_keep.index(i) for i in transform_wcs._world_keep)

    return transform_wcs, invert_xy, world_map


def wcsapi_to_celestial_frame(wcs):
    for cls, _, kwargs, *_ in wcs.world_axis_object_classes.values():
        if issubclass(cls, SkyCoord):
            return kwargs.get("frame", ICRS())
        elif issubclass(cls, BaseCoordinateFrame):
            return cls(**kwargs)


class WCSWorld2PixelTransform(CurvedTransform):
    """
    WCS transformation from world to pixel coordinates.
    """

    has_inverse = True
    frame_in = None

    def __init__(self, wcs, invert_xy=False):
        super().__init__()

        if wcs.pixel_n_dim > 2:
            raise ValueError("Only pixel_n_dim =< 2 is supported")

        self.wcs = wcs
        self.invert_xy = invert_xy

        self.frame_in = wcsapi_to_celestial_frame(wcs)

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.wcs is other.wcs
            and self.invert_xy == other.invert_xy
        )

    @property
    def input_dims(self):
        return self.wcs.world_n_dim

    def transform(self, world):
        # Convert to a list of arrays
        world = list(world.T)

        if len(world) != self.wcs.world_n_dim:
            raise ValueError(
                f"Expected {self.wcs.world_n_dim} world coordinates, got {len(world)} "
            )

        if len(world[0]) == 0:
            pixel = np.zeros((0, 2))
        else:
            pixel = self.wcs.world_to_pixel_values(*world)

        if self.invert_xy:
            pixel = pixel[::-1]

        pixel = np.array(pixel).T

        return pixel

    transform_non_affine = transform

    def inverted(self):
        """
        Return the inverse of the transform.
        """
        return WCSPixel2WorldTransform(self.wcs, invert_xy=self.invert_xy)


class WCSPixel2WorldTransform(CurvedTransform):
    """
    WCS transformation from pixel to world coordinates.
    """

    has_inverse = True

    def __init__(self, wcs, invert_xy=False):
        super().__init__()

        if wcs.pixel_n_dim > 2:
            raise ValueError("Only pixel_n_dim =< 2 is supported")

        self.wcs = wcs
        self.invert_xy = invert_xy

        self.frame_out = wcsapi_to_celestial_frame(wcs)

    def __eq__(self, other):
        return (
            isinstance(other, type(self))
            and self.wcs is other.wcs
            and self.invert_xy == other.invert_xy
        )

    @property
    def output_dims(self):
        return self.wcs.world_n_dim

    def transform(self, pixel):
        # Convert to a list of arrays
        pixel = list(pixel.T)

        if len(pixel) != self.wcs.pixel_n_dim:
            raise ValueError(
                f"Expected {self.wcs.pixel_n_dim} world coordinates, got {len(pixel)} "
            )

        if self.invert_xy:
            pixel = pixel[::-1]

        if len(pixel[0]) == 0:
            world = np.zeros((0, self.wcs.world_n_dim))
        else:
            world = self.wcs.pixel_to_world_values(*pixel)

        if self.wcs.world_n_dim == 1:
            world = [world]

        world = np.array(world).T

        return world

    transform_non_affine = transform

    def inverted(self):
        """
        Return the inverse of the transform.
        """
        return WCSWorld2PixelTransform(self.wcs, invert_xy=self.invert_xy)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from asdf.tags.core.ndarray import NDArrayType

from astropy import table
from astropy.io.misc.asdf.types import AstropyAsdfType, AstropyType


class TableType:
    """
    This class defines to_tree and from_tree methods that are used by both the
    AstropyTableType and the AsdfTableType defined below. The behavior is
    differentiated by the ``_compat`` class attribute. When ``_compat==True``,
    the behavior will conform to the table schema defined by the ASDF Standard.
    Otherwise, the behavior will conform to the custom table schema defined by
    Astropy.
    """

    _compat = False

    @classmethod
    def from_tree(cls, node, ctx):
        # This is getting meta, guys
        meta = node.get("meta", {})

        # This enables us to support files that use the table definition from
        # the ASDF Standard, rather than the custom one that Astropy defines.
        if cls._compat:
            return table.Table(node["columns"], meta=meta)

        if node.get("qtable", False):
            t = table.QTable(meta=node.get("meta", {}))
        else:
            t = table.Table(meta=node.get("meta", {}))

        for name, col in zip(node["colnames"], node["columns"]):
            t[name] = col

        return t

    @classmethod
    def to_tree(cls, data, ctx):
        columns = [data[name] for name in data.colnames]

        node = dict(columns=columns)
        # Files that use the table definition from the ASDF Standard (instead
        # of the one defined by Astropy) will not contain these fields
        if not cls._compat:
            node["colnames"] = data.colnames
            node["qtable"] = isinstance(data, table.QTable)
        if data.meta:
            node["meta"] = data.meta

        return node

    @classmethod
    def assert_equal(cls, old, new):
        assert old.meta == new.meta
        try:
            NDArrayType.assert_equal(np.array(old), np.array(new))
        except (AttributeError, TypeError, ValueError):
            for col0, col1 in zip(old, new):
                try:
                    NDArrayType.assert_equal(np.array(col0), np.array(col1))
                except (AttributeError, TypeError, ValueError):
                    assert col0 == col1


class AstropyTableType(TableType, AstropyType):
    """
    This tag class reads and writes tables that conform to the custom schema
    that is defined by Astropy (in contrast to the one that is defined by the
    ASDF Standard). The primary reason for differentiating is to enable the
    support of Astropy mixin columns, which are not supported by the ASDF
    Standard.
    """

    name = "table/table"
    types = ["astropy.table.Table"]
    requires = ["astropy"]


class AsdfTableType(TableType, AstropyAsdfType):
    """
    This tag class allows Astropy to read (and write) ASDF files that use the
    table definition that is provided by the ASDF Standard (instead of the
    custom one defined by Astropy). This is important to maintain for
    cross-compatibility.
    """

    name = "core/table"
    types = ["astropy.table.Table"]
    requires = ["astropy"]
    _compat = True


class ColumnType(AstropyAsdfType):
    name = "core/column"
    types = ["astropy.table.Column", "astropy.table.MaskedColumn"]
    requires = ["astropy"]
    handle_dynamic_subclasses = True

    @classmethod
    def from_tree(cls, node, ctx):
        data = node["data"]
        name = node["name"]
        description = node.get("description")
        unit = node.get("unit")
        meta = node.get("meta", None)

        return table.Column(
            data=data._make_array(),
            name=name,
            description=description,
            unit=unit,
            meta=meta,
        )

    @classmethod
    def to_tree(cls, data, ctx):
        node = {
            "data": data.data,
            "name": data.name,
        }
        if data.description:
            node["description"] = data.description
        if data.unit:
            node["unit"] = data.unit
        if data.meta:
            node["meta"] = data.meta

        return node

    @classmethod
    def assert_equal(cls, old, new):
        assert old.meta == new.meta
        assert old.description == new.description
        assert old.unit == new.unit

        NDArrayType.assert_equal(np.array(old), np.array(new))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

import pytest

asdf = pytest.importorskip("asdf")

import datetime

import asdf.schema as asdf_schema
import numpy as np
from asdf import AsdfFile, tagged, yamlutil
from asdf.exceptions import AsdfDeprecationWarning

with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        category=AsdfDeprecationWarning,
        message=r"asdf.tests.helpers is deprecated.*",
    )
    from asdf.tests.helpers import assert_roundtrip_tree

from astropy import time


def _flatten_combiners(schema):
    newschema = dict()

    def add_entry(path, schema, combiner):
        # TODO: Simplify?
        cursor = newschema
        for i in range(len(path)):
            part = path[i]
            if isinstance(part, int):
                cursor = cursor.setdefault("items", [])
                while len(cursor) <= part:
                    cursor.append({})
                cursor = cursor[part]
            elif part == "items":
                cursor = cursor.setdefault("items", dict())
            else:
                cursor = cursor.setdefault("properties", dict())
                if i < len(path) - 1 and isinstance(path[i + 1], int):
                    cursor = cursor.setdefault(part, [])
                else:
                    cursor = cursor.setdefault(part, dict())

        cursor.update(schema)


def test_time(tmpdir):
    time_array = time.Time(np.arange(100), format="unix")

    tree = {"large_time_array": time_array}

    assert_roundtrip_tree(tree, tmpdir)


def test_time_with_location(tmpdir):
    # See https://github.com/spacetelescope/asdf/issues/341
    from astropy import units as u
    from astropy.coordinates.earth import EarthLocation

    location = EarthLocation(x=[1, 2] * u.m, y=[3, 4] * u.m, z=[5, 6] * u.m)

    t = time.Time([1, 2], location=location, format="cxcsec")

    tree = {"time": t}

    assert_roundtrip_tree(tree, tmpdir)


def test_time_with_location_1_0_0(tmpdir):
    from astropy import units as u
    from astropy.coordinates.earth import EarthLocation

    location = EarthLocation(x=6378100 * u.m, y=0 * u.m, z=0 * u.m)

    t = time.Time("J2000.000", location=location, format="jyear_str")

    tree = {"time": t}

    # The version refers to ASDF Standard 1.0.0, which includes time-1.0.0
    assert_roundtrip_tree(tree, tmpdir, init_options={"version": "1.0.0"})


def test_isot(tmpdir):
    isot = time.Time("2000-01-01T00:00:00.000")

    tree = {"time": isot}

    assert_roundtrip_tree(tree, tmpdir)

    ff = asdf.AsdfFile(tree)
    tree = yamlutil.custom_tree_to_tagged_tree(ff.tree, ff)
    if isinstance(tree["time"], str):
        assert str(tree["time"]) == isot.value
    elif isinstance(tree["time"], dict):
        assert str(tree["time"]["value"]) == isot.value
        assert str(tree["time"]["base_format"]) == "isot"
    else:
        assert False


def test_isot_array(tmpdir):
    tree = {"time": time.Time(["2001-01-02T12:34:56", "2001-02-03T00:01:02"])}

    assert_roundtrip_tree(tree, tmpdir)


def test_time_tag():
    schema = asdf_schema.load_schema(
        "http://stsci.edu/schemas/asdf/time/time-1.1.0", resolve_references=True
    )
    schema = _flatten_combiners(schema)

    date = time.Time(datetime.datetime.now())
    tree = {"date": date}
    asdf = AsdfFile(tree=tree)
    instance = yamlutil.custom_tree_to_tagged_tree(tree["date"], asdf)

    asdf_schema.validate(instance, schema=schema)

    tag = "tag:stsci.edu:asdf/time/time-1.1.0"
    date = tagged.tag_object(tag, date)
    tree = {"date": date}
    asdf = AsdfFile(tree=tree)
    instance = yamlutil.custom_tree_to_tagged_tree(tree["date"], asdf)

    asdf_schema.validate(instance, schema=schema)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# pylint: disable=invalid-name, no-member

import numpy as np
import pytest

from astropy import units as u
from astropy.modeling.bounding_box import ModelBoundingBox
from astropy.modeling.core import fix_inputs
from astropy.modeling.fitting import (
    DogBoxLSQFitter,
    LevMarLSQFitter,
    LMLSQFitter,
    TRFLSQFitter,
)
from astropy.modeling.functional_models import (
    AiryDisk2D,
    ArcCosine1D,
    ArcSine1D,
    ArcTangent1D,
    Box1D,
    Box2D,
    Const1D,
    Const2D,
    Cosine1D,
    Disk2D,
    Ellipse2D,
    Exponential1D,
    Gaussian1D,
    Gaussian2D,
    KingProjectedAnalytic1D,
    Linear1D,
    Logarithmic1D,
    Lorentz1D,
    Moffat1D,
    Moffat2D,
    Multiply,
    Planar2D,
    RickerWavelet1D,
    RickerWavelet2D,
    Ring2D,
    Scale,
    Sersic1D,
    Sersic2D,
    Sine1D,
    Tangent1D,
    Trapezoid1D,
    TrapezoidDisk2D,
    Voigt1D,
)
from astropy.modeling.parameters import InputParameterError
from astropy.modeling.physical_models import Drude1D, Plummer1D
from astropy.modeling.polynomial import Polynomial1D, Polynomial2D
from astropy.modeling.powerlaws import (
    BrokenPowerLaw1D,
    ExponentialCutoffPowerLaw1D,
    LogParabola1D,
    PowerLaw1D,
    Schechter1D,
    SmoothlyBrokenPowerLaw1D,
)
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.compat.optional_deps import HAS_SCIPY

fitters = [LevMarLSQFitter, TRFLSQFitter, LMLSQFitter, DogBoxLSQFitter]

FUNC_MODELS_1D = [
    {
        "class": Gaussian1D,
        "parameters": {"amplitude": 3 * u.Jy, "mean": 2 * u.m, "stddev": 30 * u.cm},
        "evaluation": [(2600 * u.mm, 3 * u.Jy * np.exp(-2))],
        "bounding_box": [0.35, 3.65] * u.m,
    },
    {
        "class": Sersic1D,
        "parameters": {"amplitude": 3 * u.MJy / u.sr, "r_eff": 2 * u.arcsec, "n": 4},
        "evaluation": [(3 * u.arcsec, 1.3237148119468918 * u.MJy / u.sr)],
        "bounding_box": False,
    },
    {
        "class": Sine1D,
        "parameters": {
            "amplitude": 3 * u.km / u.s,
            "frequency": 0.25 * u.Hz,
            "phase": 0.5,
        },
        "evaluation": [(1 * u.s, -3 * u.km / u.s)],
        "bounding_box": False,
    },
    {
        "class": Cosine1D,
        "parameters": {
            "amplitude": 3 * u.km / u.s,
            "frequency": 0.25 * u.Hz,
            "phase": 0.25,
        },
        "evaluation": [(1 * u.s, -3 * u.km / u.s)],
        "bounding_box": False,
    },
    {
        "class": Tangent1D,
        "parameters": {
            "amplitude": 3 * u.km / u.s,
            "frequency": 0.125 * u.Hz,
            "phase": 0.25,
        },
        "evaluation": [(1 * u.s, -3 * u.km / u.s)],
        "bounding_box": [-4, 0] / u.Hz,
    },
    {
        "class": ArcSine1D,
        "parameters": {
            "amplitude": 3 * u.km / u.s,
            "frequency": 0.25 * u.Hz,
            "phase": 0.5,
        },
        "evaluation": [(0 * u.km / u.s, -2 * u.s)],
        "bounding_box": [-3, 3] * u.km / u.s,
    },
    {
        "class": ArcCosine1D,
        "parameters": {
            "amplitude": 3 * u.km / u.s,
            "frequency": 0.25 * u.Hz,
            "phase": 0.5,
        },
        "evaluation": [(0 * u.km / u.s, -1 * u.s)],
        "bounding_box": [-3, 3] * u.km / u.s,
    },
    {
        "class": ArcTangent1D,
        "parameters": {
            "amplitude": 3 * u.km / u.s,
            "frequency": 0.125 * u.Hz,
            "phase": 0.25,
        },
        "evaluation": [(0 * u.km / u.s, -2 * u.s)],
        "bounding_box": False,
    },
    {
        "class": Linear1D,
        "parameters": {"slope": 3 * u.km / u.s, "intercept": 5000 * u.m},
        "evaluation": [(6000 * u.ms, 23 * u.km)],
        "bounding_box": False,
    },
    {
        "class": Lorentz1D,
        "parameters": {"amplitude": 2 * u.Jy, "x_0": 505 * u.nm, "fwhm": 100 * u.AA},
        "evaluation": [(0.51 * u.micron, 1 * u.Jy)],
        "bounding_box": [255, 755] * u.nm,
    },
    {
        "class": Voigt1D,
        "parameters": {
            "amplitude_L": 2 * u.Jy,
            "x_0": 505 * u.nm,
            "fwhm_L": 100 * u.AA,
            "fwhm_G": 50 * u.AA,
        },
        "evaluation": [(0.51 * u.micron, 1.0621795524 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": Voigt1D,
        "parameters": {
            "amplitude_L": 2 * u.Jy,
            "x_0": 505 * u.nm,
            "fwhm_L": 100 * u.AA,
            "fwhm_G": 50 * u.AA,
            "method": "humlicek2",
        },
        "evaluation": [(0.51 * u.micron, 1.0621795524 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": Const1D,
        "parameters": {"amplitude": 3 * u.Jy},
        "evaluation": [(0.6 * u.micron, 3 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": Box1D,
        "parameters": {"amplitude": 3 * u.Jy, "x_0": 4.4 * u.um, "width": 1 * u.um},
        "evaluation": [(4200 * u.nm, 3 * u.Jy), (1 * u.m, 0 * u.Jy)],
        "bounding_box": [3.9, 4.9] * u.um,
    },
    {
        "class": Trapezoid1D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 4.4 * u.um,
            "width": 1 * u.um,
            "slope": 5 * u.Jy / u.um,
        },
        "evaluation": [(4200 * u.nm, 3 * u.Jy), (1 * u.m, 0 * u.Jy)],
        "bounding_box": [3.3, 5.5] * u.um,
    },
    {
        "class": RickerWavelet1D,
        "parameters": {"amplitude": 3 * u.Jy, "x_0": 4.4 * u.um, "sigma": 1e-3 * u.mm},
        "evaluation": [(1000 * u.nm, -0.09785050 * u.Jy)],
        "bounding_box": [-5.6, 14.4] * u.um,
    },
    {
        "class": Moffat1D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 4.4 * u.um,
            "gamma": 1e-3 * u.mm,
            "alpha": 1,
        },
        "evaluation": [(1000 * u.nm, 0.238853503 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": KingProjectedAnalytic1D,
        "parameters": {
            "amplitude": 1.0 * u.Msun / u.pc**2,
            "r_core": 1.0 * u.pc,
            "r_tide": 2.0 * u.pc,
        },
        "evaluation": [(0.5 * u.pc, 0.2 * u.Msun / u.pc**2)],
        "bounding_box": [0.0 * u.pc, 2.0 * u.pc],
    },
    {
        "class": Logarithmic1D,
        "parameters": {"amplitude": 5 * u.m, "tau": 2 * u.m},
        "evaluation": [(4 * u.m, 3.4657359027997265 * u.m)],
        "bounding_box": False,
    },
    {
        "class": Exponential1D,
        "parameters": {"amplitude": 5 * u.m, "tau": 2 * u.m},
        "evaluation": [(4 * u.m, 36.945280494653254 * u.m)],
        "bounding_box": False,
    },
]

SCALE_MODELS = [
    {
        "class": Scale,
        "parameters": {"factor": 2 * u.m},
        "evaluation": [(1 * u.m, 2 * u.m)],
        "bounding_box": False,
    },
    {
        "class": Multiply,
        "parameters": {"factor": 2 * u.m},
        "evaluation": [(1 * u.m / u.m, 2 * u.m)],
        "bounding_box": False,
    },
]

PHYS_MODELS_1D = [
    {
        "class": Plummer1D,
        "parameters": {"mass": 3 * u.kg, "r_plum": 0.5 * u.m},
        "evaluation": [(1 * u.m, 0.10249381 * u.kg / (u.m**3))],
        "bounding_box": False,
    },
    {
        "class": Drude1D,
        "parameters": {
            "amplitude": 1.0 * u.m,
            "x_0": 2175.0 * u.AA,
            "fwhm": 400.0 * u.AA,
        },
        "evaluation": [(2000 * u.AA, 0.5452317018423869 * u.m)],
        "bounding_box": [-17825, 22175] * u.AA,
    },
]

FUNC_MODELS_2D = [
    {
        "class": Gaussian2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_mean": 2 * u.m,
            "y_mean": 1 * u.m,
            "x_stddev": 3 * u.m,
            "y_stddev": 2 * u.m,
            "theta": 45 * u.deg,
        },
        "evaluation": [
            (412.1320343 * u.cm, 3.121320343 * u.m, 3 * u.Jy * np.exp(-0.5))
        ],
        "bounding_box": [[-13.02230366, 15.02230366], [-12.02230366, 16.02230366]]
        * u.m,
    },
    {
        "class": Const2D,
        "parameters": {"amplitude": 3 * u.Jy},
        "evaluation": [(0.6 * u.micron, 0.2 * u.m, 3 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": Disk2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "R_0": 300 * u.cm,
        },
        "evaluation": [(5.8 * u.m, 201 * u.cm, 3 * u.Jy)],
        "bounding_box": [[-1, 5], [0, 6]] * u.m,
    },
    {
        "class": TrapezoidDisk2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 1 * u.m,
            "y_0": 2 * u.m,
            "R_0": 100 * u.cm,
            "slope": 1 * u.Jy / u.m,
        },
        "evaluation": [(3.5 * u.m, 2 * u.m, 1.5 * u.Jy)],
        "bounding_box": [[-2, 6], [-3, 5]] * u.m,
    },
    {
        "class": Ellipse2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "a": 300 * u.cm,
            "b": 200 * u.cm,
            "theta": 45 * u.deg,
        },
        "evaluation": [(4 * u.m, 300 * u.cm, 3 * u.Jy)],
        "bounding_box": [
            [-0.5495097567963922, 4.549509756796392],
            [0.4504902432036073, 5.549509756796393],
        ]
        * u.m,
    },
    {
        "class": Ring2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "r_in": 2 * u.cm,
            "r_out": 2.1 * u.cm,
        },
        "evaluation": [(302.05 * u.cm, 2 * u.m + 10 * u.um, 3 * u.Jy)],
        "bounding_box": [[1.979, 2.021], [2.979, 3.021]] * u.m,
    },
    {
        "class": Box2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.s,
            "x_width": 4 * u.cm,
            "y_width": 3 * u.s,
        },
        "evaluation": [(301 * u.cm, 3 * u.s, 3 * u.Jy)],
        "bounding_box": [[0.5 * u.s, 3.5 * u.s], [2.98 * u.m, 3.02 * u.m]],
    },
    {
        "class": RickerWavelet2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "sigma": 1 * u.m,
        },
        "evaluation": [(4 * u.m, 2.5 * u.m, 0.602169107 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": AiryDisk2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "radius": 1 * u.m,
        },
        "evaluation": [(4 * u.m, 2.1 * u.m, 4.76998480e-05 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": Moffat2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 4.4 * u.um,
            "y_0": 3.5 * u.um,
            "gamma": 1e-3 * u.mm,
            "alpha": 1,
        },
        "evaluation": [(1000 * u.nm, 2 * u.um, 0.202565833 * u.Jy)],
        "bounding_box": False,
    },
    {
        "class": Sersic2D,
        "parameters": {
            "amplitude": 3 * u.MJy / u.sr,
            "x_0": 1 * u.arcsec,
            "y_0": 2 * u.arcsec,
            "r_eff": 2 * u.arcsec,
            "n": 4,
            "ellip": 0,
            "theta": 0,
        },
        "evaluation": [(3 * u.arcsec, 2.5 * u.arcsec, 2.829990489 * u.MJy / u.sr)],
        "bounding_box": False,
    },
    {
        "class": Planar2D,
        "parameters": {"slope_x": 2 * u.m, "slope_y": 3 * u.m, "intercept": 4 * u.m},
        "evaluation": [(5 * u.m / u.m, 6 * u.m / u.m, 32 * u.m)],
        "bounding_box": False,
    },
]

POWERLAW_MODELS = [
    {
        "class": PowerLaw1D,
        "parameters": {"amplitude": 5 * u.kg, "x_0": 10 * u.cm, "alpha": 1},
        "evaluation": [(1 * u.m, 500 * u.g)],
        "bounding_box": False,
    },
    {
        "class": BrokenPowerLaw1D,
        "parameters": {
            "amplitude": 5 * u.kg,
            "x_break": 10 * u.cm,
            "alpha_1": 1,
            "alpha_2": -1,
        },
        "evaluation": [(1 * u.m, 50 * u.kg), (1 * u.cm, 50 * u.kg)],
        "bounding_box": False,
    },
    {
        "class": SmoothlyBrokenPowerLaw1D,
        "parameters": {
            "amplitude": 5 * u.kg,
            "x_break": 10 * u.cm,
            "alpha_1": 1,
            "alpha_2": -1,
            "delta": 1,
        },
        "evaluation": [(1 * u.cm, 15.125 * u.kg), (1 * u.m, 15.125 * u.kg)],
        "bounding_box": False,
    },
    {
        "class": ExponentialCutoffPowerLaw1D,
        "parameters": {
            "amplitude": 5 * u.kg,
            "x_0": 10 * u.cm,
            "alpha": 1,
            "x_cutoff": 1 * u.m,
        },
        "evaluation": [(1 * u.um, 499999.5 * u.kg), (10 * u.m, 50 * np.exp(-10) * u.g)],
        "bounding_box": False,
    },
    {
        "class": LogParabola1D,
        "parameters": {"amplitude": 5 * u.kg, "x_0": 10 * u.cm, "alpha": 1, "beta": 2},
        "evaluation": [(1 * u.cm, 5 * 0.1 ** (-1 - 2 * np.log(0.1)) * u.kg)],
        "bounding_box": False,
    },
    {
        "class": Schechter1D,
        "parameters": {
            "phi_star": 1.0e-4 * (u.Mpc**-3),
            "m_star": -20.0 * u.ABmag,
            "alpha": -1.9,
        },
        "evaluation": [(-23 * u.ABmag, 1.002702276867279e-12 * (u.Mpc**-3))],
        "bounding_box": False,
    },
]

POLY_MODELS = [
    {
        "class": Polynomial1D,
        "parameters": {"degree": 2, "c0": 3 * u.one, "c1": 2 / u.m, "c2": 3 / u.m**2},
        "evaluation": [(3 * u.m, 36 * u.one)],
        "bounding_box": False,
    },
    {
        "class": Polynomial1D,
        "parameters": {
            "degree": 2,
            "c0": 3 * u.kg,
            "c1": 2 * u.kg / u.m,
            "c2": 3 * u.kg / u.m**2,
        },
        "evaluation": [(3 * u.m, 36 * u.kg)],
        "bounding_box": False,
    },
    {
        "class": Polynomial1D,
        "parameters": {"degree": 2, "c0": 3 * u.kg, "c1": 2 * u.kg, "c2": 3 * u.kg},
        "evaluation": [(3 * u.one, 36 * u.kg)],
        "bounding_box": False,
    },
    {
        "class": Polynomial2D,
        "parameters": {
            "degree": 2,
            "c0_0": 3 * u.one,
            "c1_0": 2 / u.m,
            "c2_0": 3 / u.m**2,
            "c0_1": 3 / u.s,
            "c0_2": -2 / u.s**2,
            "c1_1": 5 / u.m / u.s,
        },
        "evaluation": [(3 * u.m, 2 * u.s, 64 * u.one)],
        "bounding_box": False,
    },
    {
        "class": Polynomial2D,
        "parameters": {
            "degree": 2,
            "c0_0": 3 * u.kg,
            "c1_0": 2 * u.kg / u.m,
            "c2_0": 3 * u.kg / u.m**2,
            "c0_1": 3 * u.kg / u.s,
            "c0_2": -2 * u.kg / u.s**2,
            "c1_1": 5 * u.kg / u.m / u.s,
        },
        "evaluation": [(3 * u.m, 2 * u.s, 64 * u.kg)],
        "bounding_box": False,
    },
    {
        "class": Polynomial2D,
        "parameters": {
            "degree": 2,
            "c0_0": 3 * u.kg,
            "c1_0": 2 * u.kg,
            "c2_0": 3 * u.kg,
            "c0_1": 3 * u.kg,
            "c0_2": -2 * u.kg,
            "c1_1": 5 * u.kg,
        },
        "evaluation": [(3 * u.one, 2 * u.one, 64 * u.kg)],
        "bounding_box": False,
    },
]


MODELS = (
    FUNC_MODELS_1D
    + SCALE_MODELS
    + FUNC_MODELS_2D
    + POWERLAW_MODELS
    + PHYS_MODELS_1D
    + POLY_MODELS
)

SCIPY_MODELS = {Sersic1D, Sersic2D, AiryDisk2D}

# These models will fail fitting test, because built in fitting data
#   will produce non-finite values
NON_FINITE_LevMar_MODELS = [
    Sersic1D,
    Sersic2D,
    ArcSine1D,
    ArcCosine1D,
    PowerLaw1D,
    ExponentialCutoffPowerLaw1D,
    BrokenPowerLaw1D,
    LogParabola1D,
    Schechter1D,
]

# These models will fail the TRFLSQFitter fitting test due to non-finite
NON_FINITE_TRF_MODELS = [
    ArcSine1D,
    ArcCosine1D,
    Sersic1D,
    Sersic2D,
    PowerLaw1D,
    ExponentialCutoffPowerLaw1D,
    BrokenPowerLaw1D,
]

# These models will fail the LMLSQFitter fitting test due to non-finite
NON_FINITE_LM_MODELS = [
    Sersic1D,
    ArcSine1D,
    ArcCosine1D,
    PowerLaw1D,
    LogParabola1D,
    Schechter1D,
    ExponentialCutoffPowerLaw1D,
    BrokenPowerLaw1D,
]

# These models will fail the DogBoxLSQFitter fitting test due to non-finite
NON_FINITE_DogBox_MODELS = [
    Sersic1D,
    Sersic2D,
    ArcSine1D,
    ArcCosine1D,
    SmoothlyBrokenPowerLaw1D,
    ExponentialCutoffPowerLaw1D,
    LogParabola1D,
]


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
def test_models_evaluate_without_units(model):
    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()
    m = model["class"](**model["parameters"])
    for args in model["evaluation"]:
        if len(args) == 2:
            kwargs = dict(zip(("x", "y"), args))
        else:
            kwargs = dict(zip(("x", "y", "z"), args))
            if kwargs["x"].unit.is_equivalent(kwargs["y"].unit):
                kwargs["x"] = kwargs["x"].to(kwargs["y"].unit)
        mnu = m.without_units_for_data(**kwargs)
        args = [x.value for x in kwargs.values()]
        assert_quantity_allclose(mnu(*args[:-1]), args[-1])


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
def test_models_evaluate_with_units(model):
    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()
    m = model["class"](**model["parameters"])
    for args in model["evaluation"]:
        assert_quantity_allclose(m(*args[:-1]), args[-1])


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
def test_models_evaluate_with_units_x_array(model):
    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()

    m = model["class"](**model["parameters"])

    for args in model["evaluation"]:
        if len(args) == 2:
            x, y = args
            x_arr = u.Quantity([x, x], subok=True)
            result = m(x_arr)
            assert_quantity_allclose(result, u.Quantity([y, y], subok=True))
        else:
            x, y, z = args
            x_arr = u.Quantity([x, x])
            y_arr = u.Quantity([y, y])
            result = m(x_arr, y_arr)
            assert_quantity_allclose(result, u.Quantity([z, z]))


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
def test_models_evaluate_with_units_param_array(model):
    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()

    params = {}
    for key, value in model["parameters"].items():
        if value is None or key in ("degree", "method"):
            params[key] = value
        else:
            params[key] = np.repeat(value, 2)

    params["n_models"] = 2

    m = model["class"](**params)

    for args in model["evaluation"]:
        if len(args) == 2:
            x, y = args
            x_arr = u.Quantity([x, x], subok=True)
            result = m(x_arr)
            assert_quantity_allclose(result, u.Quantity([y, y], subok=True))
        else:
            x, y, z = args
            x_arr = u.Quantity([x, x])
            y_arr = u.Quantity([y, y])
            result = m(x_arr, y_arr)
            assert_quantity_allclose(result, u.Quantity([z, z]))

    if model["class"] == Drude1D:
        params["x_0"][-1] = 0 * u.AA
        MESSAGE = r"0 is not an allowed value for x_0"
        with pytest.raises(InputParameterError, match=MESSAGE):
            model["class"](**params)


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
def test_models_bounding_box(model):
    # In some cases, having units in parameters caused bounding_box to break,
    # so this is to ensure that it works correctly.

    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()

    m = model["class"](**model["parameters"])

    # In the following we need to explicitly test that the value is False
    # since Quantities no longer evaluate as as True
    if model["bounding_box"] is False:
        # Check that NotImplementedError is raised, so that if bounding_box is
        # implemented we remember to set bounding_box=True in the list of models
        # above
        MESSAGE = r"No bounding box is defined for this model"
        with pytest.raises(NotImplementedError, match=MESSAGE):
            m.bounding_box
    else:
        # A bounding box may have inhomogeneous units so we need to check the
        # values one by one.
        for i in range(len(model["bounding_box"])):
            bbox = m.bounding_box
            if isinstance(bbox, ModelBoundingBox):
                bbox = bbox.bounding_box()
            assert_quantity_allclose(bbox[i], model["bounding_box"][i])


@pytest.mark.parametrize("model", MODELS)
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
def test_compound_model_input_units_equivalencies_defaults(model):
    m = model["class"](**model["parameters"])

    assert m.input_units_equivalencies is None

    compound_model = m + m
    assert compound_model.inputs_map()["x"][0].input_units_equivalencies is None
    fixed_input_model = fix_inputs(compound_model, {"x": 1})

    assert fixed_input_model.input_units_equivalencies is None

    compound_model = m - m
    assert compound_model.inputs_map()["x"][0].input_units_equivalencies is None
    fixed_input_model = fix_inputs(compound_model, {"x": 1})

    assert fixed_input_model.input_units_equivalencies is None

    compound_model = m & m
    assert compound_model.inputs_map()["x1"][0].input_units_equivalencies is None
    fixed_input_model = fix_inputs(compound_model, {"x0": 1})
    assert fixed_input_model.inputs_map()["x1"][0].input_units_equivalencies is None

    assert fixed_input_model.input_units_equivalencies is None

    if m.n_outputs == m.n_inputs:
        compound_model = m | m
        assert compound_model.inputs_map()["x"][0].input_units_equivalencies is None
        fixed_input_model = fix_inputs(compound_model, {"x": 1})

        assert fixed_input_model.input_units_equivalencies is None


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
@pytest.mark.filterwarnings(r"ignore:.*:RuntimeWarning")
@pytest.mark.filterwarnings(r"ignore:Model is linear in parameters.*")
@pytest.mark.filterwarnings(r"ignore:The fit may be unsuccessful.*")
@pytest.mark.filterwarnings(r"ignore:humlicek2 has been deprecated since .*")
@pytest.mark.parametrize("model", MODELS)
@pytest.mark.parametrize("fitter", fitters)
def test_models_fitting(model, fitter):
    fitter = fitter()

    bad_voigt = model["class"] == Voigt1D and ("method" not in model["parameters"])
    if (
        (
            isinstance(fitter, LevMarLSQFitter)
            and model["class"] in NON_FINITE_LevMar_MODELS
        )
        or (
            isinstance(fitter, TRFLSQFitter)
            and (model["class"] in NON_FINITE_TRF_MODELS or bad_voigt)
        )
        or (
            isinstance(fitter, LMLSQFitter)
            and (model["class"] in NON_FINITE_LM_MODELS or bad_voigt)
        )
        or (
            isinstance(fitter, DogBoxLSQFitter)
            and model["class"] in NON_FINITE_DogBox_MODELS
        )
    ):
        return

    m = model["class"](**model["parameters"])
    if len(model["evaluation"][0]) == 2:
        x = np.linspace(1, 3, 100) * model["evaluation"][0][0].unit
        y = np.exp(-x.value**2) * model["evaluation"][0][1].unit
        args = [x, y]
    else:
        x = np.linspace(1, 3, 100) * model["evaluation"][0][0].unit
        y = np.linspace(1, 3, 100) * model["evaluation"][0][1].unit
        z = np.exp(-x.value**2 - y.value**2) * model["evaluation"][0][2].unit
        args = [x, y, z]

    # Test that the model fits even if it has units on parameters
    m_new = fitter(m, *args)

    # Check that units have been put back correctly
    for param_name in m.param_names:
        par_bef = getattr(m, param_name)
        par_aft = getattr(m_new, param_name)
        if par_bef.unit is None:
            # If the parameter used to not have a unit then had a radian unit
            # for example, then we should allow that
            assert par_aft.unit is None or par_aft.unit is u.rad
        else:
            assert par_aft.unit.is_equivalent(par_bef.unit)


unit_mismatch_models = [
    {
        "class": Gaussian2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_mean": 2 * u.m,
            "y_mean": 1 * u.m,
            "x_stddev": 3 * u.m,
            "y_stddev": 2 * u.m,
            "theta": 45 * u.deg,
        },
        "evaluation": [
            (412.1320343 * u.cm, 3.121320343 * u.K, 3 * u.Jy * np.exp(-0.5)),
            (412.1320343 * u.K, 3.121320343 * u.m, 3 * u.Jy * np.exp(-0.5)),
        ],
        "bounding_box": [[-14.18257445, 16.18257445], [-10.75693665, 14.75693665]]
        * u.m,
    },
    {
        "class": Ellipse2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "a": 300 * u.cm,
            "b": 200 * u.cm,
            "theta": 45 * u.deg,
        },
        "evaluation": [(4 * u.m, 300 * u.K, 3 * u.Jy), (4 * u.K, 300 * u.cm, 3 * u.Jy)],
        "bounding_box": [[-0.76046808, 4.76046808], [0.68055697, 5.31944302]] * u.m,
    },
    {
        "class": Disk2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "R_0": 300 * u.cm,
        },
        "evaluation": [
            (5.8 * u.m, 201 * u.K, 3 * u.Jy),
            (5.8 * u.K, 201 * u.cm, 3 * u.Jy),
        ],
        "bounding_box": [[-1, 5], [0, 6]] * u.m,
    },
    {
        "class": Ring2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "r_in": 2 * u.cm,
            "r_out": 2.1 * u.cm,
        },
        "evaluation": [
            (302.05 * u.cm, 2 * u.K + 10 * u.K, 3 * u.Jy),
            (302.05 * u.K, 2 * u.m + 10 * u.um, 3 * u.Jy),
        ],
        "bounding_box": [[1.979, 2.021], [2.979, 3.021]] * u.m,
    },
    {
        "class": TrapezoidDisk2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 1 * u.m,
            "y_0": 2 * u.m,
            "R_0": 100 * u.cm,
            "slope": 1 * u.Jy / u.m,
        },
        "evaluation": [
            (3.5 * u.m, 2 * u.K, 1.5 * u.Jy),
            (3.5 * u.K, 2 * u.m, 1.5 * u.Jy),
        ],
        "bounding_box": [[-2, 6], [-3, 5]] * u.m,
    },
    {
        "class": RickerWavelet2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "sigma": 1 * u.m,
        },
        "evaluation": [
            (4 * u.m, 2.5 * u.K, 0.602169107 * u.Jy),
            (4 * u.K, 2.5 * u.m, 0.602169107 * u.Jy),
        ],
        "bounding_box": False,
    },
    {
        "class": AiryDisk2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "radius": 1 * u.m,
        },
        "evaluation": [
            (4 * u.m, 2.1 * u.K, 4.76998480e-05 * u.Jy),
            (4 * u.K, 2.1 * u.m, 4.76998480e-05 * u.Jy),
        ],
        "bounding_box": False,
    },
    {
        "class": Moffat2D,
        "parameters": {
            "amplitude": 3 * u.Jy,
            "x_0": 4.4 * u.um,
            "y_0": 3.5 * u.um,
            "gamma": 1e-3 * u.mm,
            "alpha": 1,
        },
        "evaluation": [
            (1000 * u.nm, 2 * u.K, 0.202565833 * u.Jy),
            (1000 * u.K, 2 * u.um, 0.202565833 * u.Jy),
        ],
        "bounding_box": False,
    },
    {
        "class": Sersic2D,
        "parameters": {
            "amplitude": 3 * u.MJy / u.sr,
            "x_0": 1 * u.arcsec,
            "y_0": 2 * u.arcsec,
            "r_eff": 2 * u.arcsec,
            "n": 4,
            "ellip": 0,
            "theta": 0,
        },
        "evaluation": [
            (3 * u.arcsec, 2.5 * u.m, 2.829990489 * u.MJy / u.sr),
            (3 * u.m, 2.5 * u.arcsec, 2.829990489 * u.MJy / u.sr),
        ],
        "bounding_box": False,
    },
]


@pytest.mark.parametrize("model", unit_mismatch_models)
def test_input_unit_mismatch_error(model):
    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()

    MESSAGE = "Units of 'x' and 'y' inputs should match"

    m = model["class"](**model["parameters"])

    for args in model["evaluation"]:
        if len(args) == 2:
            kwargs = dict(zip(("x", "y"), args))
        else:
            kwargs = dict(zip(("x", "y", "z"), args))
            if kwargs["x"].unit.is_equivalent(kwargs["y"].unit):
                kwargs["x"] = kwargs["x"].to(kwargs["y"].unit)
        with pytest.raises(u.UnitsError, match=MESSAGE):
            m.without_units_for_data(**kwargs)


mag_models = [
    {
        "class": Const1D,
        "parameters": {"amplitude": 3 * u.ABmag},
        "evaluation": [(0.6 * u.ABmag, 3 * u.ABmag)],
    },
    {
        "class": Const1D,
        "parameters": {"amplitude": 3 * u.ABmag},
        "evaluation": [(0.6 * u.mag, 3 * u.ABmag)],
    },
    {
        "class": Const1D,
        "parameters": {"amplitude": 3 * u.mag},
        "evaluation": [(0.6 * u.ABmag, 3 * u.mag)],
    },
    {
        "class": Const1D,
        "parameters": {"amplitude": 3 * u.mag},
        "evaluation": [(0.6 * u.mag, 3 * u.mag)],
    },
    {
        "class": Const2D,
        "parameters": {"amplitude": 3 * u.ABmag},
        "evaluation": [(0.6 * u.micron, 0.2 * u.m, 3 * u.ABmag)],
    },
    {
        "class": Ellipse2D,
        "parameters": {
            "amplitude": 3 * u.ABmag,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "a": 300 * u.cm,
            "b": 200 * u.cm,
            "theta": 45 * u.deg,
        },
        "evaluation": [(4 * u.m, 300 * u.cm, 3 * u.ABmag)],
    },
    {
        "class": Disk2D,
        "parameters": {
            "amplitude": 3 * u.ABmag,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "R_0": 300 * u.cm,
        },
        "evaluation": [(5.8 * u.m, 201 * u.cm, 3 * u.ABmag)],
    },
    {
        "class": Ring2D,
        "parameters": {
            "amplitude": 3 * u.ABmag,
            "x_0": 3 * u.m,
            "y_0": 2 * u.m,
            "r_in": 2 * u.cm,
            "r_out": 2.1 * u.cm,
        },
        "evaluation": [(302.05 * u.cm, 2 * u.m + 10 * u.um, 3 * u.ABmag)],
    },
    {
        "class": Box2D,
        "parameters": {
            "amplitude": 3 * u.ABmag,
            "x_0": 3 * u.m,
            "y_0": 2 * u.s,
            "x_width": 4 * u.cm,
            "y_width": 3 * u.s,
        },
        "evaluation": [(301 * u.cm, 3 * u.s, 3 * u.ABmag)],
    },
    {
        "class": SmoothlyBrokenPowerLaw1D,
        "parameters": {
            "amplitude": 5 * u.ABmag,
            "x_break": 10 * u.cm,
            "alpha_1": 1,
            "alpha_2": -1,
            "delta": 1,
        },
        "evaluation": [(1 * u.cm, 15.125 * u.ABmag), (1 * u.m, 15.125 * u.ABmag)],
    },
    {
        "class": Box1D,
        "parameters": {"amplitude": 3 * u.ABmag, "x_0": 4.4 * u.um, "width": 1 * u.um},
        "evaluation": [(4200 * u.nm, 3 * u.ABmag), (1 * u.m, 0 * u.ABmag)],
        "bounding_box": [3.9, 4.9] * u.um,
    },
    {
        "class": Schechter1D,
        "parameters": {
            "phi_star": 1.0e-4 * (u.Mpc**-3),
            "m_star": -20.0 * u.ABmag,
            "alpha": -1.9,
        },
        "evaluation": [(-23 * u.ABmag, 1.002702276867279e-12 * (u.Mpc**-3))],
    },
    {
        "class": Schechter1D,
        "parameters": {
            "phi_star": 1.0e-4 * (u.Mpc**-3),
            "m_star": -20.0 * u.mag,
            "alpha": -1.9,
        },
        "evaluation": [(-23 * u.mag, 1.002702276867279e-12 * (u.Mpc**-3))],
    },
]


@pytest.mark.parametrize("model", mag_models)
def test_models_evaluate_magunits(model):
    if not HAS_SCIPY and model["class"] in SCIPY_MODELS:
        pytest.skip()

    m = model["class"](**model["parameters"])
    for args in model["evaluation"]:
        assert_quantity_allclose(m(*args[:-1]), args[-1])


def test_Schechter1D_errors():
    # Non magnitude units are bad
    model = Schechter1D(
        phi_star=1.0e-4 * (u.Mpc**-3), m_star=-20.0 * u.km, alpha=-1.9
    )
    MESSAGE = r"The units of magnitude and m_star must be a magnitude"
    with pytest.raises(u.UnitsError, match=MESSAGE):
        model(-23 * u.km)

    # Differing magnitude systems are bad
    model = Schechter1D(
        phi_star=1.0e-4 * (u.Mpc**-3), m_star=-20.0 * u.ABmag, alpha=-1.9
    )
    MESSAGE = (
        r".*: Units of input 'x', .*, could not be converted to required input units"
        r" of .*"
    )
    with pytest.raises(u.UnitsError, match=MESSAGE):
        model(-23 * u.STmag)

    # Differing magnitude systems are bad
    model = Schechter1D(
        phi_star=1.0e-4 * (u.Mpc**-3), m_star=-20.0 * u.ABmag, alpha=-1.9
    )
    with pytest.raises(u.UnitsError, match=MESSAGE):
        model(-23 * u.mag)



from math import ceil

import numpy as np


def _n_tiles(data_shape, tile_shape):
    return [int(ceil(d / t)) for d, t in zip(data_shape, tile_shape)]


def _iter_array_tiles(
    data_shape, tile_shape, first_tile_index=None, last_tile_index=None
):
    ndim = len(tile_shape)

    n_tiles = _n_tiles(data_shape, tile_shape)

    if first_tile_index is None:
        first_tile_index = (0,) * ndim

    if last_tile_index is None:
        last_tile_index = tuple(n - 1 for n in n_tiles)

    tile_index = list(first_tile_index)

    while True:
        tile_slices = tuple(
            slice(
                (tile_index[idx] - first_tile_index[idx]) * tile_shape[idx],
                (tile_index[idx] - first_tile_index[idx] + 1) * tile_shape[idx],
            )
            for idx in range(ndim)
        )

        row_index = tile_index[0]
        for dim in range(1, ndim):
            row_index = tile_index[dim] + row_index * n_tiles[dim]

        yield row_index, tile_slices

        tile_index[-1] += 1

        for idx in range(ndim - 1, 0, -1):
            if tile_index[idx] > last_tile_index[idx]:
                tile_index[idx] = first_tile_index[idx]
                tile_index[idx - 1] += 1

        if tile_index[0] > last_tile_index[0]:
            break


def _tile_shape(header):
    return np.array(
        [header[f"ZTILE{idx}"] for idx in range(header["ZNAXIS"], 0, -1)], dtype=int
    )


def _data_shape(header):
    return np.array(
        [header[f"ZNAXIS{idx}"] for idx in range(header["ZNAXIS"], 0, -1)], dtype=int
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from numpy.testing import assert_array_equal

from astropy.io.misc.asdf.tags.transform.basic import TransformType
from astropy.modeling import physical_models

from . import _parameter_to_value

__all__ = ["BlackBody", "Drude1DType", "Plummer1DType"]


class BlackBody(TransformType):
    name = "transform/blackbody"
    version = "1.0.0"
    types = ["astropy.modeling.physical_models.BlackBody"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return physical_models.BlackBody(
            scale=node["scale"], temperature=node["temperature"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "scale": _parameter_to_value(model.scale),
            "temperature": _parameter_to_value(model.temperature),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, physical_models.BlackBody) and isinstance(
            b, physical_models.BlackBody
        )
        assert_array_equal(a.scale, b.scale)
        assert_array_equal(a.temperature, b.temperature)


class Drude1DType(TransformType):
    name = "transform/drude1d"
    version = "1.0.0"
    types = ["astropy.modeling.physical_models.Drude1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return physical_models.Drude1D(
            amplitude=node["amplitude"], x_0=node["x_0"], fwhm=node["fwhm"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "amplitude": _parameter_to_value(model.amplitude),
            "x_0": _parameter_to_value(model.x_0),
            "fwhm": _parameter_to_value(model.fwhm),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, physical_models.Drude1D) and isinstance(
            b, physical_models.Drude1D
        )
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.fwhm, b.fwhm)


class Plummer1DType(TransformType):
    name = "transform/plummer1d"
    version = "1.0.0"
    types = ["astropy.modeling.physical_models.Plummer1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return physical_models.Plummer1D(mass=node["mass"], r_plum=node["r_plum"])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {
            "mass": _parameter_to_value(model.mass),
            "r_plum": _parameter_to_value(model.r_plum),
        }
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert isinstance(a, physical_models.Plummer1D) and isinstance(
            b, physical_models.Plummer1D
        )
        assert_array_equal(a.mass, b.mass)
        assert_array_equal(a.r_plum, b.r_plum)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Define a constant to know if the entry points are installed, since this impacts
# whether we can run the tests.

from importlib.metadata import entry_points

import pytest

# TODO: Exclusively use select when Python minversion is 3.10
eps = entry_points()
if hasattr(eps, "select"):
    ep = [entry.name for entry in eps.select(group="asdf_extensions")]
else:
    ep = [entry.name for entry in eps.get("asdf_extensions", [])]
ASDF_ENTRY_INSTALLED = "astropy" in ep and "astropy-asdf" in ep

del entry_points, eps, ep

if not ASDF_ENTRY_INSTALLED:
    pytest.skip(
        "The astropy asdf entry points are not installed", allow_module_level=True
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import FK4, ICRS, Galactic, Longitude, SkyCoord

asdf = pytest.importorskip("asdf")

from asdf.exceptions import AsdfDeprecationWarning

with warnings.catch_warnings():
    warnings.filterwarnings(
        "ignore",
        category=AsdfDeprecationWarning,
        message=r"asdf.tests.helpers is deprecated.*",
    )
    from asdf.tests.helpers import assert_roundtrip_tree

# These tests are cribbed directly from the Examples section of
# https://docs.astropy.org/en/stable/api/astropy.coordinates.SkyCoord.html


def test_scalar_skycoord(tmpdir):
    c = SkyCoord(10, 20, unit="deg")  # defaults to ICRS frame
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


def test_vector_skycoord(tmpdir):
    c = SkyCoord([1, 2, 3], [-30, 45, 8], frame="icrs", unit="deg")  # 3 coords
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


def test_skycoord_fk4(tmpdir):
    coords = ["1:12:43.2 +1:12:43", "1 12 43.2 +1 12 43"]
    c = SkyCoord(coords, frame=FK4, unit=(u.deg, u.hourangle), obstime="J1992.21")
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


@pytest.mark.parametrize(
    "coord",
    [
        SkyCoord("1h12m43.2s +1d12m43s", frame=Galactic),  # Units from string
        SkyCoord(frame="galactic", l="1h12m43.2s", b="+1d12m43s"),
    ],
)
def test_skycoord_galactic(coord, tmpdir):
    tree = dict(coord=coord)
    assert_roundtrip_tree(tree, tmpdir)


def test_skycoord_ra_dec(tmpdir):
    ra = Longitude([1, 2, 3], unit=u.deg)  # Could also use Angle
    dec = np.array([4.5, 5.2, 6.3]) * u.deg  # Astropy Quantity
    c = SkyCoord(ra, dec, frame="icrs")
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)

    c = SkyCoord(frame=ICRS, ra=ra, dec=dec, obstime="2001-01-02T12:34:56")
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


def test_skycoord_override_defaults(tmpdir):
    c = FK4(1 * u.deg, 2 * u.deg)  # Uses defaults for obstime, equinox
    c = SkyCoord(c, obstime="J2010.11", equinox="B1965")  # Override defaults
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


def test_skycoord_cartesian(tmpdir):
    c = SkyCoord(
        w=0, u=1, v=2, unit="kpc", frame="galactic", representation_type="cartesian"
    )
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


def test_skycoord_vector_frames(tmpdir):
    c = SkyCoord([ICRS(ra=1 * u.deg, dec=2 * u.deg), ICRS(ra=3 * u.deg, dec=4 * u.deg)])
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


@pytest.mark.xfail(reason="Velocities are not properly serialized yet")
def test_skycoord_radial_velocity(tmpdir):
    c = SkyCoord(ra=1 * u.deg, dec=2 * u.deg, radial_velocity=10 * u.km / u.s)
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


@pytest.mark.xfail(reason="Velocities are not properly serialized yet")
def test_skycoord_proper_motion(tmpdir):
    c = SkyCoord(
        ra=1 * u.deg,
        dec=2 * u.deg,
        pm_ra_cosdec=2 * u.mas / u.yr,
        pm_dec=1 * u.mas / u.yr,
    )
    tree = dict(coord=c)
    assert_roundtrip_tree(tree, tmpdir)


@pytest.mark.skip(reason="Apparent loss of precision during serialization")
def test_skycoord_extra_attribute(tmpdir):
    sc = SkyCoord(10 * u.deg, 20 * u.deg, equinox="2011-01-01T00:00", frame="fk4")
    tree = dict(coord=sc.transform_to("icrs"))

    def check_asdf(asdffile):
        assert hasattr(asdffile["coord"], "equinox")

    assert_roundtrip_tree(tree, tmpdir, asdf_check_func=check_asdf)


def test_skycoord_2d_obstime(tmpdir):
    sc = (
        SkyCoord(
            [1, 2],
            [3, 4],
            [5, 6],
            unit="deg,deg,m",
            frame="fk4",
            obstime=["J1990.5", "J1991.5"],
        ),
    )
    tree = dict(coord=sc)
    assert_roundtrip_tree(tree, tmpdir)
